# ftp_down.py
# Dependencies: ftplib, pandas

import io
import os.path
import ftplib
import pandas as pd
from customsol_pkg.errors import FTPConnectionError, FTPDownloadError, FTPEncodingError


def ftp_download(hostname, username, password, filename_filter, download_directory='.'):
    """
    ----------------------------------------------------------------
    Fetch information using FTP and save file to local machine.
    ----------------------------------------------------------------
    Parameters:
        hostname : string
            FTP host name
        username : string
            FTP login username
        password : string
            FTP login password
        filename_filter : string
            Name of file to download from FTP
        download_directory: string
            Direct to download file to. CWD is default
    Returns:
       NONE
    Side Effects:
        Writes and deletes feed file to local machine (in same dir as script)
            Output file will have same name as FTP file
            (ie. same as filename_filter parameter)
    ----------------------------------------------------------------
    """
    try:
        # Login to FTP
        print(f"Initializing FTP Download from {hostname}...")
        ftp = ftplib.FTP(hostname)
    except OSError:
        raise FTPConnectionError('Unable to establish connect to host')
    try:
        print("Attempting FTP Login...")
        ftp.login(username, password)
        print("FTP Login Successful!")
    except ftplib.error_perm as e:
        raise FTPConnectionError('Incorrect Login.')

    try:
        ftp.encoding = "utf-8"
    except Exception as e:
        raise FTPEncodingError('Unable to enforce UTF-8 Encoding.')

    try:
        print(f"Starting Download of {filename_filter}...")
        # Downloads a copy of the FTP file, saving in current dir with same name as filename_filter
        ftp.retrbinary("RETR " + filename_filter, open(os.path.join(download_directory, filename_filter), 'wb').write)
        print("File Downloaded Successfully!\n")
        ftp.quit()

    except ftplib.error_perm as err:
        raise FTPDownloadError('Unable to open file.')



# up_ftp.py
# Dependencies: ftplib

import os
import ftplib
from customsol_pkg.errors import FTPConnectionError, FTPUploadError, FTPEncodingError

def ftp_upload_file(hostname, username, password, file_name):
    """
    -------------------------------------------------------
    Uploads a given file to FTP
    -------------------------------------------------------
    Parameters:
        hostname : string
            FTP host name
        username : string
            FTP login username
        password : string
            FTP login password
        file_name : string
            Name of file in dir that will be uploaded
        upload_name : string
            !! MUST INCLUDE EXTENSION !!
            (Optional) Specifies name of FTP upload
                ie. can be uploaded with different name than local file
    ------------------------------------------------------
    """
    try:
        # Login to FTP
        print(f"Initializing FTP Upload to {hostname}...")
        ftp = ftplib.FTP(hostname)
    except OSError:
        raise FTPConnectionError('Unable to establish connect to host')
    try:
        print("Attempting FTP Login...")
        ftp.login(username, password)
        print("FTP Login Successful!")
    except ftplib.error_perm as e:
        raise FTPConnectionError('Incorrect Login.')
    try:
        print(f"Starting Upload of {file_name}...")

        # Store file to FTP with appropriate upload name
        ftp.storbinary("STOR " + (file_name), open(file_name, 'rb'))
        print("File Upload Successful!")

        ftp.quit()

    except ftplib.all_errors as err:
        raise FTPUploadError("ERROR: Could not upload to FTP.")



# dir_clean.py

import os
import os.path

def dir_clean_cwd(folder_path=""):
    """
    -------------------------------------------------------
    Removes all excel and csv files from current dir
    -------------------------------------------------------
    Parameters:
        folder_path : string
            (Optional) Path to directory containing files to delete
            If not specified, deletes from current directory
    Returns:
       NONE
    Side Effects:
        Deletes all .xlsx and .csv files from local dir
    ------------------------------------------------------
    """

    # Cwd (Default)
    files_in_directory = os.listdir(os.getcwd())

    # Chosen directory to clean
    if not (folder_path == ""):
        files_in_directory = os.listdir(os.path.normpath(folder_path))

    # Remove all .xlsx files
    filtered_files = [file for file in files_in_directory if file.endswith(".xlsx")]
    for file in filtered_files:
        path_to_file = os.path.join('', file)
        os.remove(path_to_file)

    # Remove all .csv files
    filtered_files = [file for file in files_in_directory if file.endswith(".csv")]
    for file in filtered_files:
        path_to_file = os.path.join('', file)
        os.remove(path_to_file)

    print("\nTemporary files deleted from local machine!")



# slack_funcs.py

from slack import WebClient
from slack.errors import SlackApiError
import sys

SLACK_BOT_TOKEN = 'xoxb-4950217216-1580739024241-KE4CsYwOouHLxmpuEmk8BYJ4'
client = WebClient(token=SLACK_BOT_TOKEN)
def slack_send_message(channels, messages):
    """
    -------------------------------------------------------
    Sends slack message(s) to specified channel(s)
    Use: slack_send_message(channels, messages))
    -------------------------------------------------------
        channels: list
            contains list of channels to send messages to via slack
        messages: list
            contains list of messages to send
    Returns:
       NONE
    ------------------------------------------------------
    """
    sys.dont_write_bytecode = True
    for channel in channels:
        for message in messages:
            response = client.chat_postMessage(channel=channel, text=message)
    return


def slack_send_file(channels, files, titles):
    """
    -------------------------------------------------------
    Sends file(s) to specified channel(s)
    Use: slack_send_file(channels, files, titles)
    -------------------------------------------------------
        channels: list
            contains list of channels to send messages to via slack
        files: list
            contains list of files to send
        titles: list
            contains list of titles of files; titles shoud correspond to same
            order of items in files list.
    Returns:
       NONE
    ------------------------------------------------------
    """
    for channel in channels:
        i = 0
        for file in files:
            title = titles[i]
            client.files_upload(channels=channel, file=file, title=title)
            i += 1
    return



# supp_create.py
# Dependencies: pandas

import pandas as pd
import numpy as np
from customsol_pkg.errors import DataError

def create_supplement(df, dealership_col, vin_col, stock_col, trim_col, jato_col='',
    chrmstyle_col='', discount_col='', misc1_col='', misc2_col=''):

    if str(type(df)) != '<class \'pandas.core.frame.DataFrame\'>':
        raise TypeError('Must pass a DataFrame Object.')

    # List of supplement columns
    supp_columns = ["Dealership", "VIN", "Stock", "Jato ID", "ChrmStyleID", "Trim", "Discount", "Misc 1", "Misc 2"]

    # List of column variables
    fill_data = ['dealership_col', 'vin_col', 'stock_col', 'trim_col', 'jato_col',
        'chrmstyle_col', 'discount_col', 'misc1_col', 'misc2_col']

    # Mapping column variables to column names
    match_dict = {
    'dealership_col': 'Dealership',
    'vin_col': 'VIN',
    'stock_col': 'Stock',
    'trim_col': 'Trim',
    'jato_col': 'Jato ID',
    'chrmstyle_col': 'ChrmStyleID',
    'discount_col': 'Discount',
    'misc1_col': 'Misc 1',
    'misc2_col': 'Misc 2'
    }

    # Create list of passed variables, empty variables, and mapping those variables to header names
    header_list = []
    missing_header_list = []
    header_map = []

    # Iterate through local variables to determine which variables are passed and which are empty strings
    for arg in locals().items():
        if arg[0] in fill_data:
            if arg[1] != '':
                header_list.append(arg[1])
                header_map.append(arg)
            else:
                missing_header_list.append(arg[0])

    # Create new dataframe using only GIVEN parameterslem
    try:
        supp_df = df[header_list].copy()
    except KeyError:
        raise KeyError('Passed columns names do not match columns in passed DataFrame')

    # Rename columns using mapping list and respective dict entries
    for i in range(supp_df.shape[1]):
        supp_df = supp_df.rename(columns={header_map[i][1]:match_dict.get(header_map[i][0])})

    # Add columns to data frame from list of empty headers
    i = 0
    for entry in missing_header_list:
        supp_df[match_dict.get(entry)] = ''
        i+=1

    # Change order of columns to match supplement file
    supp_df = supp_df[supp_columns]

    return supp_df



class CustomError(Exception):
    pass


class DataError(CustomError):
    pass


class FTPConnectionError(CustomError):
    pass


class FTPDownloadError(CustomError):
    pass


class FTPUploadError(CustomError):
    pass


class FTPEncodingError(CustomError):
    pass


class RedashAPIError(CustomError):
    pass


class DuplicateFileError(CustomError):
    pass



# down_redash.py
# Dependencies - requests, pandas

import os
import requests
import time
import pandas as pd
import datetime
from customsol_pkg.errors import RedashAPIError

def redash_download(query_id, api_key, region, file_name, params={}):
    """
    -------------------------------------------------------
    Download query response to local dir as file_name
    -------------------------------------------------------
    Parameters:
        query_id : string
            ID of Redash Query
        api_key : string
            User API key
        region : string
            Country for Redash (ie. CA or US)
        file_name : string
            File will be downloaded and saved locally as csv with this name
            !! NO EXTENSION !!
        params : Dictionary
            (Optional) Dictionary of query params -- Default = {} (empty)

    Raises:
        ValueError: Region code is Invalid
        RedashAPIError: Unable to connect to Redash API or requested query
    ------------------------------------------------------
    """

    # Create request header using provided API key
    header = {'Authorization': f'Key {api_key}'}

    # Use provided region to generate a job url and pass param
    job_url = ''
    pass_param = ''
    redash_csv = ''

    if region.lower() == 'ca':
        job_url = 'https://redash.motocommerce.ca/api/jobs/{}'
        pass_param = f'https://redash.motocommerce.ca/api/queries/{query_id}/refresh'
        redash_csv = f'https://redash.motocommerce.ca/api/queries/{query_id}/results.csv?api_key={api_key}'

    elif region.lower() == 'us' or region.lower() == 'usa':
        job_url = 'https://redash.motocommerce.com/api/jobs/{}'
        pass_param = f'https://redash.motocommerce.com/api/queries/{query_id}/refresh'
        redash_csv = f'https://redash.motocommerce.com/api/queries/{query_id}/results.csv?api_key={api_key}'

    else:
        raise ValueError('Invalid region code passed. Use CA for Canada and US for USA.')

    # Download data from Redash query


    print(f'Exporting Query - ID: {query_id}...')

    # Set Up Requests Object
    s = requests.Session()
    s.headers.update(header)
    print('Fetching report using Redash API...')
    # Post the Refresh with the New Parameters
    response = s.post(pass_param, params=params)
    print(f'Redash Response Status Code: {response.status_code}')

    if response.status_code == 404:
        raise RedashAPIError('Invalid API Key')

    if response.status_code == 400:
        raise RedashAPIError('Invalid Query ID')

    # Wait for Refresh Job to Complete
    job = response.json()['job']

    while job['status'] not in (3, 4):
        response = s.get(job_url.format(job['id']))
        job = response.json()['job']
        time.sleep(1)

    if job['status'] == 3:
        param_id = job['query_result_id']
    else:
        param_id = None

    # Return Updated Query
    response = s.get(redash_csv.format(param_id))

    # Export data to .csv file
    text_file = open(str(file_name + '.csv'), 'w', newline='', encoding='utf-8')
    n = text_file.write(response.text)
    text_file.close()

    print(f'Downloaded {file_name}\n')



    return



# ftp_cleanse.py
# Dependencies: pandas

import pandas as pd

def ftp_cleanse_file(ftp_df):

    dummy_fill_val = "AAAAAAAAAAAAAAAAA"

    # Create a copy of the downloaded dataframe
    dummy_df = ftp_df.copy()

    # Remove all but the first row in downloaded dataframe
    dummy_df = dummy_df.iloc[:1]

    # Find which columns are filled and which are not
    filled_cols = []
    for col in dummy_df.columns:
        isFilled = not dummy_df[col].isna().any()
        filled_cols.append(isFilled)

    # Replace all filled columns with dummy value
    index_counter = 0

    for isFilled in filled_cols:
        if isFilled:
            dummy_df[dummy_df.columns[index_counter]] = [dummy_fill_val]
        index_counter += 1

    # Upload the dummy value, returning the result of upload
    return dummy_df
